/* arm specific support for Elf loading and relocation.
   Copyright 1996, 1997 Linux International.

   Contributed by wms <woody@corelcomputer.com>

   This file is part of the Linux modutils.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ident "$Id: obj_arm.c,v 1.1 1997/09/10 22:20:02 wms Exp $"

#include <string.h>
#include <assert.h>

#include <module.h>
#include <obj.h>
#include <util.h>


/*======================================================================*/

struct arm_plt_entry
{
  struct arm_plt_entry *next;
  ElfW(Addr) r_info;
  int offset;
  int inited;
};

struct arm_file
{
  struct obj_file file;
  struct obj_section *plt;
};

struct arm_symbol
{
  struct obj_symbol sym;
  struct arm_plt_entry *plt_entries;
};

/*======================================================================*/

struct obj_file *
arch_new_file (void)
{
  struct arm_file *f;

  f = xmalloc(sizeof(struct arm_file));
  f->plt = NULL;
  return &f->file;
}

struct obj_section *
arch_new_section (void)
{
  return xmalloc(sizeof(struct obj_section));
}

struct obj_symbol *
arch_new_symbol (void)
{
  struct arm_symbol *p;

  p = xmalloc(sizeof(struct arm_symbol));
  p->plt_entries = NULL;
  return &p->sym;
}

/*======================================================================*/

enum obj_reloc
arch_apply_relocation (struct obj_file *f,
		       struct obj_section *targsec,
		       struct obj_section *symsec,
		       struct obj_symbol *sym,
		       Elf32_Rel *rel,
		       Elf32_Addr v)
{
  struct arm_file *afile = (struct arm_file *)f;
  struct arm_symbol *asym  = (struct arm_symbol *)sym;

  Elf32_Addr *loc = (Elf32_Addr *)(targsec->contents + rel->r_offset);
  Elf32_Addr dot = targsec->header.sh_addr + rel->r_offset;

  struct arm_plt_entry *pe;
#define instruction long unsigned
  instruction *ip;

  
  
  enum obj_reloc ret = obj_reloc_ok;

  switch (ELF32_R_TYPE(rel->r_info))
    {
    case R_ARM_NONE:
      break;

    case R_ARM_ABS32:
      *loc += v;
      break;

    case R_ARM_PC24:
      /* find the plt entry and initialize it if necessary */
      assert(asym != NULL);
      for (pe = asym->plt_entries; pe != NULL && pe->r_info != rel->r_info; )
	pe = pe->next;
      assert(pe != NULL);
      if (!pe->inited)
	{
	  ip = (instruction *) (afile->plt->contents + pe->offset);
	  ip[0] = 0xe51ff004;			/* ldr pc,[pc] */
	  ip[1] = v;	 			/* sym@ */
	  pe->inited = 1;
	}

      v -= dot;
      if ((int)v < -0x02000000 || (int)v >= 0x02000000)
	{
	  /* go via the plt */
	  v = afile->plt->header.sh_addr + pe->offset - dot;
	}
      if (v & 3)
	ret = obj_reloc_dangerous;
	
      v -= 8;		//ARM specific - adjust PC
      v >>= 2;		//ARM specific - offset in a branch is in dwords
      *loc = (*loc & ~0x00ffffff) | (v & 0x00ffffff);
      break;

    default:
      ret = obj_reloc_unhandled;
      break;
    }

  return ret;
}



int
arch_create_got (struct obj_file *f)
{
  struct arm_file *pf = (struct arm_file *) f;
  int i, offset;
  struct obj_section *sec, *syms, *strs;
  ElfW(Rel) *rel, *relend;
  ElfW(Sym) *symtab, *extsym;
  const char *strtab, *name;
  struct arm_symbol *intsym;
  struct arm_plt_entry *pe;

  offset = 0;
  for (i = 0; i < f->header.e_shnum; ++i)
    {
      sec = f->sections[i];
      if (sec->header.sh_type != SHT_RELM)
	continue;
      syms = f->sections[sec->header.sh_link];
      strs = f->sections[syms->header.sh_link];

      rel = (ElfW(RelM) *) sec->contents;
      relend = rel + (sec->header.sh_size / sizeof(ElfW(RelM)));
      symtab = (ElfW(Sym) *) syms->contents;
      strtab = (const char *) strs->contents;

      for (; rel < relend; ++rel)
	{
	  if (ELF32_R_TYPE(rel->r_info) != R_ARM_PC24)
	    continue;
	  extsym = &symtab[ELF32_R_SYM(rel->r_info)];
	  if (extsym->st_name)
	    name = strtab + extsym->st_name;
	  else
	    name = f->sections[extsym->st_shndx]->name;
	  intsym = (struct arm_symbol *) obj_find_symbol(f, name);

	  for (pe = intsym->plt_entries; pe != NULL; pe = pe->next)
	    if (pe->r_info == rel->r_info)
	      break;
	  if (pe == NULL)
	    {
	      pe = xmalloc(sizeof(struct arm_plt_entry));
	      pe->next = intsym->plt_entries;
	      pe->r_info = rel->r_info;
	      pe->offset = offset;
	      pe->inited = 0;
	      intsym->plt_entries = pe;
	      offset += 8;
	    }
	}
    }

  pf->plt = obj_create_alloced_section(f, ".plt", 8, offset);

  return 1;
}


int
arch_init_module (struct obj_file *f, struct new_module *mod)
{
  return 1;
}
